#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any, Dict

import mesh.log as log
import mesh.telemetry as telemetry
import mesh.tool as tool
from mesh.cause import MeshCode, Codeable, MeshException
from mesh.context import Mesh
from mesh.macro import spi
from mesh.mpc.digest import Digest
from mesh.mpc.filter import Filter, Invoker, Invocation, PROVIDER
from mesh.prsim import Metadata
from mesh.prsim.context import RunMode


@spi(name="provider", pattern=PROVIDER, priority=0x7fffff00)
class ProviderFilter(Filter):

    def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        if not attachments:
            attachments = {}
        Mesh.context().decode(attachments)

        digest = Digest()
        try:
            ret = invoker.run(invocation)
            digest.write("P", MeshCode.SUCCESS.get_code())
            return ret
        except BaseException as e:
            if isinstance(e, Codeable):
                digest.write("P", e.get_code())
            else:
                digest.write("P", MeshCode.SYSTEM_ERROR.get_code())
            if isinstance(e, MeshException):
                log.error(f"{digest.trace_id},{Mesh.context().get_urn()},{e.get_message}")
            raise e


@spi(name="telemetryProvider", pattern=PROVIDER, priority=0x7ffffff0)
class TelemetryProviderFilter(Filter):
    """
    recover telemetry context via mesh context for tracing
    """

    def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        if not RunMode.TRACE.match(Mesh.context().get_run_mode()):
            return invoker.run(invocation)

        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        if not telemetry.if_rebuild_span():
            return invoker.run(invocation)
        trace_id = attachments.get(Metadata.MESH_TRACE_ID.key(), tool.new_trace_id())
        with telemetry.build_via_remote(attachments, invocation.get_urn().string()):
            current_span = telemetry.get_current_span()
            current_span.set_attribute('mesh-trace-id', trace_id)
            return invoker.run(invocation)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.sidecar.sidecar import open_sidecar

__all__ = (
    "open_sidecar"
)


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import os
import platform
import subprocess
import tempfile
from abc import ABC, abstractmethod
from importlib import resources
from typing import Dict, Callable, Any, IO

import mesh.tool as tool
from mesh.macro import spi, mpi, ServiceLoader
from mesh.mpc import Mesh
from mesh.prsim import Network


@spi(name="mesh")
class Sidecar(ABC):

    @abstractmethod
    @mpi("${mesh.name}.sidecar.allocate")
    def allocate(self, inst_id: str) -> Dict[str, str]:
        """
        :param inst_id:
        :return:
        """
        pass

    @abstractmethod
    @mpi("${mesh.name}.sidecar.finalize")
    def finalize(self, inst_id: str):
        """
        :param inst_id:
        :return:
        """
        pass


class Selector:

    def __init__(self, inst_id: str):
        self.inst_id = inst_id
        self.addrs: Dict[str, str] = {}

    def __enter__(self) -> Dict[str, str]:
        self.addrs = Mesh.context_safe(lambda: self.invoke(lambda: self.sidecar().allocate(self.inst_id)))
        return self.addrs

    def __exit__(self, exc_type, exc_val, exc_tb):
        Mesh.context_safe(lambda: self.invoke(lambda: self.sidecar().finalize(self.inst_id)))

    @staticmethod
    def invoke(routine: Callable) -> Any:
        network = ServiceLoader.load(Network).get_default()
        network.get_environ()
        Mesh.context().set_attribute(Mesh.REMOTE_NAME, "sidecar")
        Mesh.context().set_attribute(Mesh.REMOTE, os.environ.get("MESH_SIDECAR_ADDRESS", "127.0.0.1:7204"))
        return routine()

    @mpi
    def sidecar(self) -> Sidecar:
        pass


class SidecarDaemon:

    def __init__(self):
        uname = platform.system().lower()
        suffix = "" if uname != "windows" else ".exe"
        name = f"sidecar-{uname}-amd64{suffix}"
        body = resources.read_binary(__package__, name)
        fd, executable = tempfile.mkstemp(name)
        os.close(fd)
        os.chmod(executable, os.stat(executable).st_mode | 0o755)
        with open(file=executable, mode='wb+') as f:
            f.write(body)
            self.executable = executable
            self.running = False

    def open(self, inst_id: str, stdout: IO = None) -> Selector:
        if not self.running:
            try:
                subprocess.Popen(self.executable, stdout=stdout, stderr=stdout, env={
                    "MESH_NAME": "sidecar",
                    "MESH_ADDRESS": tool.get_mesh_address().string(),
                    "MESH_RUNTIME": f"{tool.get_mesh_runtime().hostname}:7204"})
                self.running = True
            except subprocess.CalledProcessError as e:
                raise e

        return Selector(inst_id)


sidecar = SidecarDaemon()


def open_sidecar(inst_id: str, stdout: IO = None) -> Selector:
    return sidecar.open(inst_id, stdout)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import socket
import unittest

from mesh.sidecar import open_sidecar


class TestSidecar(unittest.TestCase):

    def test_tcp(self):
        with open_sidecar('LX0000010008760') as selector:
            conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            vs = selector.get('libpsi').split(":")
            conn.connect((vs[0], int(vs[1])))
            for i in range(1000):
                conn.send(f"Hello {i}\n".encode())
            while True:
                data = conn.recv(1024)
                if not data:
                    break
                print(data.decode())

            conn.close()


if __name__ == '__main__':
    unittest.main()
